//-----------------------------------------------------------------------------
//
//                      XMM Science Analysis System (SAS)
//                     (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	  sao2xcolorcod.cc	--- convert SAOimage color tables to xcolorcod format
//
//	DESCRIPTION:
//		Convert a given SAOimage ASCII colortable to xcolorcod format.
//		Usage is:
//			sao2xcolorcod outsetname
//		SAOimage color table is read from stdin; xcolorcod color table is
//		written to dataset <outsetname>.ds in CWD.
//
//	CHANGE HISTORY:
//		UL, 2002-02-19: 1.0.0: - first version
//
//	AUTHORS:
//		Uwe Lammers (UL), Wed Feb 19 15:38:04 MET 2003
//
// $Header: /sasbuild/repositories/sasdev/sas/packages/xcolorcod/colortable/sao2xcolorcod.cc,v 1.2 2009/09/01 15:25:38 aibarra Exp $
//
#include <algorithm>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <string>

#include "stringutil.h"
#include "../src/ColorTable.h"

using namespace std;

class SAOColorTableFile {
public:
	SAOColorTableFile(istream &is) : _is(&is), _lineNr(0) {}

	vector<double> scan(vector<ColorTable> &ctabs) const {
		string line;
		vector<double> levels;
		while (getline(*_is, line) && _is->good()) {
			_lineNr++;
			line = trim(line, true);
			if (line.empty() || line[0] == '#')
				//  empty or comment line
				continue;
			unsigned ci;
			if (line == "RED:")
				ci = 0;
			else if (line == "GREEN:")
				ci = 1;
			else if (line == "BLUE:")
				ci = 2;
			else
				//	skip this line
				continue;

			_lineNr++;
			getline(*_is, line);
			if (line.empty() || !_is->good())
				err();
			vector<CPoint> cp = cbranch(line);
			if (cp.size() < 2)
				err();

			for (vector<CPoint>::const_iterator i = cp.begin();
				 i != cp.end(); ++i) {
				Color color(0., 0., 0.);	// black
				color.component(ci) += i->second;
				//cout<<"Adding color: "<<color<<" to level: "<<i->first<<endl;
				levels.push_back(i->first);
				ctabs[ci][i->first] += color;
			}
		}

		return levels;
	}

private:
	istream	*_is;
	mutable unsigned	_lineNr;

	typedef pair<double, double> CPoint;
	vector<CPoint> cbranch(const string &line) const {
		vector<CPoint> cv;
		vector<string> v = breakup(line, "()");
		for (unsigned n=0; n<v.size(); ++n) {
			vector<string> s = breakup(v[n], ",");
			if (s.size() != 2)
				err();
			double lev = toNumber<double>(s[0]),
				   col = toNumber<double>(s[1]);
			/* AI gcc4 porting
			 *  error: suggest explicit braces to avoid ambiguous ‘else’
			 */
			if (!cv.empty())
			{
				if (lev == cv.back().first)
				{
					// duplicated level value
					if (col == cv.back().second) {
						//	duplicate color intensity - skip this
						cout<<"Duplicate entry in line "
							<<_lineNr<<" element "<<n
							<<" - ignoring this\n";
						continue;
					} else {
						//	add epsilon to level
						cout<<"Duplicated level in line "
							<<_lineNr<<" element "<<n
							<<" - adding epsilon to value\n";
						lev += 1.e-6;
					}
				}
			}
			cv.push_back(make_pair(lev, col));
		}

		if (cv.size() == 1 && cv.front().first == 0.
						   && cv.front().second == 0.)
			cv.push_back(make_pair(1., 0.));

		return cv;
	}

	void err() const {
		cerr<<"SAOimage color table syntax error in line #"<<_lineNr<<endl;
		exit(1);
	}
};

int main(int argc, char **argv)
{
	if (argc != 2) {
		cerr<<"Usage: "<<argv[0]<<" outsetname\n";
		return 1;
	}

	vector<ColorTable> comptab(3);
	SAOColorTableFile ctf(cin);
	vector<double> levels = ctf.scan(comptab);
	//	construct resulting color table
	ColorTable cTab;
	sort(levels.begin(), levels.end());
	vector<double>::iterator p = unique(levels.begin(), levels.end());
	levels.erase(p, levels.end());
	for (vector<double>::const_iterator i = levels.begin(); i != levels.end();
		 ++i)
		cTab[*i] = comptab[0](*i) + comptab[1](*i) + comptab[2](*i);

	cTab.toDataSet(argv[1]);
	return 0;
} // main()
