//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	Color: A class representing a color.
//
//	DESCRIPTION:
//		A class that represents a color composed of a red/green/blue
//		component.
//
//	CHANGE HISTORY:
//		MB, 1999-??-??: 1.0.0: - first version
//
//	AUTHORS:
//		M. Beijersbergen (MB)
//		U. Lammers (UL), Mon Jul 17 13:46:09 MET DST 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/xcolorcod/src/Color.h,v 1.7 2003/06/16 08:53:05 gvacanti Exp $
//
#ifndef COLOR_H
#define COLOR_H

#include<cassert>
#include <iostream>
#include "errstr.h"

class Color {
public:
	Color(double r=0., double g=0., double b=0.) : red(r), green(g), blue(b) {}

	Color& operator+=(const Color& c) {
		red += c.red;
		saturationCheck(red, 0);
		green += c.green;
		saturationCheck(green, 1);
		blue += c.blue;
		saturationCheck(blue, 2);

		return *this;
	}

	Color& operator-=(const Color& c) {
		red -= c.red; green -= c.green; blue -= c.blue;
		return *this;
	}

	Color& operator*=(double f) {
		red *= f; green *= f; blue *= f;
		return *this;
	}

	bool operator==(const Color &c) const {
		return red==c.red && green==c.green && blue==c.blue;
	}

	bool operator!=(const Color &c) const {
		return !operator==(c);
	}

	bool isWhite() const {
		return operator==(Color::white);
	}

	bool isBlack() const {
		return operator==(Color::black);
	}

	inline double &component(unsigned n) {
		if (n==0)
			return red;
		if (n==1)
			return green;
		return blue;
	}

	double red, green, blue;

	static Color black;
	static Color white;

private:
	void saturationCheck(double &c, unsigned short i) {
		static const char *cols[] = { "red", "green", "blue" };
		assert(i<=2);
		if (c > 1.) {
			errstr<<"Saturation of "<<cols[i]<<" component reached ("
				  <<c<<')'<<Sas::warning("ColorSaturation");
			c = 1.;
		}
	}
};


inline Color operator+(const Color& c1, const Color& c2) {
	Color r = c1;
	r += c2;
	return r;
}

inline Color operator-(const Color& c1, const Color& c2) {
	Color r = c1;
	r -= c2;
	return r;
}

inline Color operator*(double f, const Color& c) {
	Color r = c;
	r *= f;
	return r;
}

inline std::ostream& operator<<(std::ostream& os, const Color& c)
{
	return os<<'('<<c.red<<", "<<c.green<<", "<<c.blue<<')';
}

#endif
