//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	ColorImage.cc: Implementation of ColorImage methods
//
//	DESCRIPTION:
//		Methods to construct/modify/access ColorImage.
//
//	CHANGE HISTORY:
//		MB, 1999-??-??: 1.0.0: - first version
//		UL, 2000-07-19: 2.0.0: - various changed related to 1.4 overhaul
//							   - added methods to write RGB image as
//								 data cube in primary array
//
//	AUTHORS:
//		M. Beijersbergen (MB)
//		U. Lammers (UL), Mon Jul 17 13:46:09 MET DST 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/xcolorcod/src/ColorImage.cc,v 1.13 2013/08/02 12:09:09 aibarra Exp $
//
#include <cmath>

#include "ColorImage.h"
#include "Dal.h"
#include "param.h"

using namespace std;

ColorImage::ColorImage(int nx, int ny) : _pixels(0)
{
	size(nx, ny);
} // ColorImage::ColorImage()


void ColorImage::size(int nx, int ny)
{
	delete _pixels;
	_pixels = new Color[nx*ny];
	_nx = nx;
	_ny = ny;

} // ColorImage::size()


Color& ColorImage::operator()(int ix, int iy)
{
	if (ix < 0 || ix >= _nx ||
		iy < 0 || iy >= _ny)
		return Color::black;

	return _pixels[iy*_nx+ix];
} // ColorImage::operator()


const Color& ColorImage::operator()(int ix, int iy) const
{
	if (ix < 0 || ix >= _nx ||
		iy < 0 || iy >= _ny)
		return Color::black;

	return _pixels[iy*_nx+ix];

} // ColorImage::operator()


double ColorImage::max() const
{
	double cmax = 0;
	for (int iy = 0; iy < _ny; ++iy)
		for (int ix = 0; ix < _nx; ++ix) {
			Color c = (*this)(ix,iy);
			if (c.red > cmax) cmax = c.red;
			if (c.green > cmax) cmax = c.green;
			if (c.blue > cmax) cmax = c.blue;
		}

	return cmax;
} // ColorImage::max()


void ColorImage::normalize(double newmax)
{
	double cmax = max();
	if (cmax <= 0)
		return;

	double factor = newmax/cmax;

	for (int iy = 0; iy < _ny; ++iy)
		for (int ix = 0; ix < _nx; ++ix)
				(*this)(ix,iy) *= factor;

} // ColorImage::normalize()


void ColorImage::add(ColorImage* image,vector<double> weight)
{
	cout <<"weight " << weight[1] << " " << weight[3] << " "
			<< weight[5]<< endl;
	for (int iy = 0; iy < _ny; ++iy)
		for (int ix = 0; ix < _nx; ++ix) {
			Color& c = (*this)(ix,iy);
			Color& c2 = (*image)(ix,iy);

			if (c.red >= 0)
				c.red += c2.red * (weight[1]);
			if (c.green >= 0)
				c.green += c2.green* (weight[3]);
			if (c.blue >= 0)
				c.blue += c2.blue* (weight[5]);
		}
}

void ColorImage::applyWeight(vector<double> weight)
{
	cout <<"weight0 " << weight[1] << " " << weight[3] << " "
			<< weight[5]<< endl;
	for (int iy = 0; iy < _ny; ++iy)
		for (int ix = 0; ix < _nx; ++ix) {
			Color& c = (*this)(ix,iy);

			if (c.red >= 0)
				c.red = c.red * weight[1];
			if (c.green >= 0)
				c.green = c.green * weight[3];
			if (c.blue >= 0)
				c.blue = c.blue * weight[5];
		}
}

void ColorImage::logScale(double binSize, double scalMin, double scalMax)
{
	double cmax = max();
	if (cmax <= 0)
		return;

	cmax = (scalMax*((binSize/1200.0)*(binSize/1200.0)))/1000.;
	double cmin = (scalMin*((binSize/1200.0)*(binSize/1200.0)))/1000.;
	double logcmax = ::log10(cmax);
	double logcmin = ::log10(cmin);

	for (int iy = 0; iy < _ny; ++iy)
		for (int ix = 0; ix < _nx; ++ix) {
			Color& c = (*this)(ix,iy);

			if (c.red >= 0) {
				double v;
				if (::log10(c.red) >= logcmax)
					v = 255.;
				else if (::log10(c.red) <= logcmin)
					v = 0;
				else
					v = ((::log10(c.red) - logcmin) / (logcmax - logcmin ) ) * 255.0;
				c.red = (v >= 0 ? v : 0);
			}
			if (c.green >= 0) {
				double v;
				if (::log10(c.green) >= logcmax)
					v = 255.;
				else if (::log10(c.green) <= logcmin)
					v = 0;
				else
					v = ((::log10(c.green)- logcmin) / (logcmax - logcmin ) ) * 255.0;

				c.green = (v >= 0 ? v : 0);
			}
			if (c.blue >= 0) {
				double v;

				if (::log10(c.blue) >= logcmax)
					v = 255.;
				else if (::log10(c.blue) <= logcmin)
					v = 0;
				else
					v = ((::log10(c.blue)- logcmin) / (logcmax - logcmin ) ) * 255.0;

				c.blue = (v >= 0 ? v : 0);
			}
		}
} // ColorImage::log()



void ColorImage::log(double decades)
{
	double cmax = max();
	if (cmax <= 0)
		return;

	double logcmax = ::log10(cmax);

	for (int iy = 0; iy < _ny; ++iy)
		for (int ix = 0; ix < _nx; ++ix) {
			Color& c = (*this)(ix,iy);

			if (c.red >= 0) {
				double	v = ::log10(c.red) - logcmax + decades;
				c.red = (v >= 0 ? v : 0);
			}
			if (c.green >= 0) {
				double v = ::log10(c.green) - logcmax + decades;
				c.green = (v >= 0 ? v : 0);
			}
			if (c.blue >= 0) {
				double v = ::log10(c.blue) - logcmax + decades;
				c.blue = (v >= 0 ? v : 0);
			}  
		}
} // ColorImage::log()


void ColorImage::writePPM(ostream& os) const
{
	os<<"P6"<<endl;
	// The X is the Y and viceversa....
	//The first component of the array is the Y and 
	//the second in the X.
	//See TableBasedPseudoRGBImage::toTrueColorImage()
	os<<_ny<<' '<<_nx<<endl;
	os<<max()<<endl;
	unsigned char pixels[_nx*_ny*3];
	unsigned char *ptr = &pixels[0];
	Color *col = _pixels;	
	for (int ix = 0; ix < _nx; ++ix) 
	  for (int iy = 0; iy < _ny; ++iy)	
	     {
		#if 0
			unsigned short r = static_cast<unsigned short>(col->red),
			g = static_cast<unsigned short>(col->green),
			b = static_cast<unsigned short>(col->blue);
	        cout<<iy<<'/'<<ix<<" : "<<r<<' '<<g<<' '<<b<<endl;
		#endif
			//Needed to rotate the image -180 degrees.
			*ptr++ = static_cast<unsigned char>((*this)(_nx-ix,iy).red);
			*ptr++ = static_cast<unsigned char>((*this)(_nx-ix,iy).green);
			*ptr++ = static_cast<unsigned char>((*this)(_nx-ix,iy).blue);

		//	*ptr++ = static_cast<unsigned char>(col->red);
		//	*ptr++ = static_cast<unsigned char>(col->green);
		//      *ptr++ = static_cast<unsigned char>(col->blue);
			++col;
		}
	os.write(reinterpret_cast<const char *>(pixels), sizeof(pixels));
} // ColorImage::writePPM()


void ColorImage::writeAsRGBarrays(DataSet *set)
{
	longVec	size;
	size.push_back(_nx);
	size.push_back(_ny);

	//	write red array
	ArrayData *rd = set->addArray("RED", Dal::Int8, size)->data();
	int8 *r = rd->int8Data();
	Color *col = _pixels;
	for (int iy=0; iy<_ny; ++iy)
		for (int ix=0; ix<_nx; ++ix)
			*r++ = static_cast<int8>((col++)->red+.5);

	//	write green array
	ArrayData *gd = set->addArray("GREEN", Dal::Int8, size)->data();
	int8 *g = gd->int8Data();
	col = _pixels;
	for (int iy=0; iy<_ny; ++iy)
		for (int ix=0; ix<_nx; ++ix) {
			*g++ = static_cast<int8>(col->green+.5);
			++col;
	}

	//	write blue array
	ArrayData *bd = set->addArray("BLUE", Dal::Int8, size)->data();
	int8 *b = bd->int8Data();
	col = _pixels;
	for (int iy=0; iy<_ny; ++iy)
		for (int ix=0; ix<_nx; ++ix) {
			*b++ = static_cast<int8>(col->blue+.5);
			++col;
	}

	delete rd;
	delete gd;
	delete bd;
} // ColorImage::writeAsRGBarrays()


void ColorImage::writeAsDataCube(DataSet *set,std::vector<std::string> evlistFileName,
		double scalmin, double scalmax)
{
	//Extract WCS information from original table
	//std::vector<std::string> evlistFileName = stringListParameter("tablelist");

	//Extract the ObsID information an close file
	Table *inTab = dataSetServer->open(evlistFileName[0], Dal::Read)->table("EVENTS");
	std::vector<real64>binSize ;
	binSize.push_back(80.);
	binSize.push_back(80.);



	*set->addAttribute( "HDUCLASS",
			"Format conforms to OGIP/GSFC conventions") = "OGIP";
	*set->addAttribute( "HDUCLAS1","File contains an Image") = "IMAGE";

	*set->addAttribute( "HDUCLAS2","Gross Image") = "TOTAL";

	*set->addAttribute( "HDUVERS1", "Version of format") = "1.1.0";
	std::vector<Column *> coord;
	Column *xCol, *yCol= 0;
	xCol = inTab->column("X");
	yCol = inTab->column("Y");
	coord.push_back(xCol);
	coord.push_back(yCol);

	std::vector<real64> xLimits(2), yLimits(2);
	xLimits.push_back( xCol->attribute("TLMIN")->asReal());
	xLimits.push_back(xCol->attribute("TLMAX")->asReal());
	yLimits.push_back(yCol->attribute("TLMIN")->asReal());
	yLimits.push_back(yCol->attribute("TLMAX")->asReal());

	if( xLimits[0] != floor(xLimits[0]) || xLimits[1] != floor(xLimits[1]) ){
		errstr << "X ranges must be integers for integer-valued columns."<<Sas::error("XRangeVal");
	}
	xLimits[0] -= 0.5;
	xLimits[1] += 0.5;



	if( yLimits[0] != floor(yLimits[0]) || yLimits[1] != floor(yLimits[1]) ){
		errstr << "Y ranges must be integers for integer-valued columns."<<Sas::error("YRangeVal");
	}
	yLimits[0] -= 0.5;
	yLimits[1] += 0.5;

	std::vector<real64> coordMin;
	coordMin.push_back(xLimits[0]);
	coordMin.push_back(yLimits[0]);

	std::vector<real64> wcsBin;
	wcsBin.push_back(binSize[1]);
	wcsBin.push_back(binSize[0]);

	for(unsigned ico=0; ico<2; ico++){
		std::stringstream typeName, pixName, valName, deltName, unitName, ltvName, ltmName;
	    typeName << "CTYPE"<<ico+1;
	    pixName << "CRPIX"<<ico+1;
	    valName << "CRVAL"<<ico+1;
	    deltName << "CDELT"<<ico+1;
	    unitName << "CUNIT"<<ico+1;
	    ltvName << "LTV"<<ico+1;
	    ltmName << "LTM"<<ico+1<<"_"<<ico+1;
	    std::stringstream typeNamel, pixNamel, valNamel, deltNamel, unitNamel;
	    typeNamel << "CTYPE"<<ico+1<<"L"<<std::ends;
	    pixNamel << "CRPIX"<<ico+1<<"L"<<std::ends;
	    valNamel << "CRVAL"<<ico+1<<"L"<<std::ends;
	    deltNamel << "CDELT"<<ico+1<<"L"<<std::ends;
	    unitNamel << "CUNIT"<<ico+1<<"L"<<std::ends;

	    set->addAttribute(coord[ico]->attribute("TCTYP"))->rename(typeName.str());

	    real64 pixVal = coord[ico]->attribute("TCRPX")->asReal();
	    *set->addAttribute(pixName.str(), "WCS reference pixel") = (pixVal-coordMin[ico])/wcsBin[ico] + 0.5;

	    set->addAttribute(coord[ico]->attribute("TCRVL"))->rename(valName.str());

	    if(coord[ico]->hasAttribute("TCUNI")) set->addAttribute(coord[ico]->attribute("TCUNI"))->rename(unitName.str());

	    real64 deltVal = coord[ico]->attribute("TCDLT")->asReal();

	    *set->addAttribute(deltName.str(), "WCS pixel size") = deltVal*wcsBin[ico];
	}
	if (inTab->hasAttribute("INSTRUME"))
		*set->addAttribute("INSTRUME", "Instrument name") = this->getInstrument();
	if (inTab->hasAttribute("FILTER"))
		*set->addAttribute("FILTER", "Filter ID") = this->getFilter();
	if (inTab->hasAttribute("OBJECT"))
		*set->addAttribute("OBJECT", "Name of observed object") = inTab->attribute("OBJECT")->asString();
	if (inTab->hasAttribute("OBSERVER"))
		*set->addAttribute("OBSERVER", "Name of PI") = inTab->attribute("OBSERVER")->asString();
	if (inTab->hasAttribute("DATE-OBS"))
		*set->addAttribute("DATE-OBS", "tart Time (UTC) of exposure") = inTab->attribute("DATE-OBS")->asString();

	*set->addAttribute("EXPOSURE", "Exposure time") = this->getExposure();
    *set->addAttribute("SCALMIN", "min value for log scaling") = scalmin;
    *set->addAttribute("SCALMAX", "max value for log scaling") = scalmax;




	dataSetServer->close(inTab->parent());





	// RGB image
	longVec	size;
	size.push_back(3);
	size.push_back(_nx);
	size.push_back(_ny);

	if (_trueColor == false)
		normalize(255);
	ArrayData *ad = set->addArray("PRIMARY", Dal::Real32, size)->data();
	real32 *rgb = ad->real32Data();
	longVec posVec(3);
	for (int iy=0; iy<_ny; ++iy)
		for (int ix=0; ix<_nx; ++ix) {
			posVec[2] = iy; posVec[1] = ix;

			posVec[0] = 0; rgb[ad->position(posVec)] =
									static_cast<real32>((*this)(ix, iy).red);
			posVec[0] = 1; rgb[ad->position(posVec)] =
									static_cast<real32>((*this)(ix, iy).green);
			posVec[0] = 2; rgb[ad->position(posVec)] =
									static_cast<real32>((*this)(ix, iy).blue);
		}
	delete ad;


} // ColorImage::writeAsDataCube()


void ColorImage::writeDataSet(const std::string &name,
		std::vector<std::string> evlistFileName,
		bool asCube, double scalmin, double scalmax)
{
	DataSet *set = dataSetServer->open(name, DataSet::Create);

	if (asCube)
		writeAsDataCube(set,evlistFileName,scalmin,scalmax);
	else
		writeAsRGBarrays(set);

	dataSetServer->close(set);

} // ColorImage::writeDataSet()


void ColorImage::setTrueColor(bool trueColor)
{
	_trueColor = trueColor;
}
bool ColorImage::isTrueColor()
{
	return _trueColor;
}

ColorImage::~ColorImage()
{
	delete[] _pixels;
} // ColorImage::~ColorImage()
