//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	TableBasedPseudoRGBImage : RGB image in the form of a table
//
//	DESCRIPTION:
//		A class that represents an RGB image in the form of a table
//		with three columns "RED"/"GREEN"/"BLUE"
//
//	CHANGE HISTORY:
//		UL, 2000-07-17: 1.0.0: - first version
//		UL, 2000-10-30: 1.1.0: - fixed problem with forgotten data set
//								 close (caused test harness to fail in a weird
//								 manner in lowmem)
//		UL, 2000-11-09: 1.2.0: - No! Above fix only cured the low-mem failure
//								 by pure chance. The problem was really
//								 trying to read a newly created data set
//								 before it is closed:
//									newset = dataSetServer->open(...)
//										populate(newset)
//										read(newset)
//									dataSetServer->close(newset)
//								 Unknown if this is a Dal bug or feature.
//								 New set is now closed before it is being
//								 used
//	    AI, 2009-05-05: 1.22  New methods added to deal with the new true color
//							  procedure defined by Pedro Rodriguez.
//
//	AUTHOR:
//		U. Lammers (UL), Mon Jul 17 13:46:09 MET DST 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/xcolorcod/src/TableBasedPseudoRGBImage.cc,v 1.12 2013/08/02 12:09:09 aibarra Exp $
//
#include <cstdio>	//	tmpnam()

#include "TableBasedPseudoRGBImage.h"
#include "ConvertableData.h"

#include "metatask.h"
#include "evselect.h"

TableBasedPseudoRGBImage::TableBasedPseudoRGBImage(
			const string &table,
			const string &xColName,
			const string &yColName,
			const string &eColName,
			MinMaxDeterminationScheme minMaxScheme,
			double minVal, double maxVal,
			const LinearlyInterpolatedPointsCurve<double,Color> &colorCurve)
	: _tempTableName("tempcolorset.ds:COLORTAB")
{
	Table *inTab = dataSetServer->open(table, Dal::Read)->table(table);
	unsigned N = inTab->rows();
	const Column *x = inTab->column(xColName),
				 *y = inTab->column(yColName);

	_livetime = inTab->attribute("LIVETIME")->asReal();
	//etruecolor assume that the input event files have the 
	//same astrometry origin. This fix has been introduced after 
	//runnning etruecolor with mosaicing images.
	//_ra_pnt = inTab->attribute("RA_PNT")->asReal();
	//_dec_pnt = inTab->attribute("DEC_PNT")->asReal();

	vector<double> weightVec;
	if (inTab->hasColumn("WEIGHT")){
		Column *wcol = inTab->column("WEIGHT");
		ColumnData *wColDat = wcol->data();
		real32 *wDat = wColDat->real32Data();

		for (unsigned n=0; n<N; ++n)
			weightVec.push_back(wDat[n]);

		delete wColDat;
	}
	else
		for (unsigned n=0; n<N; ++n)
			weightVec.push_back(1.0);

	Table *tempTable  = dataSetServer->open(_tempTableName, Dal::Create)
												->addTable(_tempTableName, N);

	tempTable->add(x);
	tempTable->add(y);


	_cdr = tempTable->addColumn("RED", Dal::Real32)->data(),
	_cdg = tempTable->addColumn("GREEN", Dal::Real32)->data(),
	_cdb = tempTable->addColumn("BLUE", Dal::Real32)->data();
	_r = _cdr->real32Data();
	_g = _cdg->real32Data();
	_b = _cdb->real32Data();

	for (unsigned n=0; n<N; ++n)
		_r[n] = _g[n] = _b[n] = 0.f;

	Column *eCol = inTab->column(eColName);
	ConvertableData cde(eCol->data());
	if (minMaxScheme == LEGAL) {
		minVal = eCol->attribute("TLMIN")->asReal();
		maxVal = eCol->attribute("TLMAX")->asReal();
	} else if (minMaxScheme == ACTUAL) {
		minVal = 1.e30;
		maxVal = -1.e30;
		for (unsigned n=0; n<N; ++n) {
			double v = cde.asDouble(n);
			minVal = std::min(v, minVal);
			maxVal = std::max(v, maxVal);
		}
	}

	double d=maxVal-minVal;
	if (d==0.)
		errstr<<"Minimum and maximum data value are equal - this value is: "
			  <<maxVal<<Sas::error("MinMaxEqual");
	for (unsigned n=0; n<N; ++n) {
		double v = std::max(std::min(cde.asDouble(n), maxVal), minVal);
		Color col = colorCurve((v-minVal)/d);
		//cout<< "PI " <<cde.asDouble(n) << " Color for pixel val: "<<(v-minVal)/d<<": "<<col<<endl;
		col *= weightVec[n];
		_r[n] += col.red;
		_g[n] += col.green;
		_b[n] += col.blue;
	}

	delete _cdr;
	delete _cdg;
	delete _cdb;

	dataSetServer->close(tempTable->parent());
	dataSetServer->close(inTab->parent());
} // TableBasedPseudoRGBImage::TableBasedPseudoRGBImage()

//
//	convert to a ColorImage using evselect
//
ColorImage *TableBasedPseudoRGBImage::toColorImage() const
{
	ColorImage *colImage = 0;

	vector<string> colNames(3);
	colNames[0] = "RED";
	colNames[1] = "GREEN";
	colNames[2] = "BLUE";

	ParameterInfo *p = Info::Task::current()->parameters();
	string xcolumn(p->parameter("xcolumn")->stringValue()),
		   ycolumn(p->parameter("ycolumn")->stringValue()),
		   imagebinning(p->parameter("imagebinning")->stringValue());
	double ximagebinsize(p->parameter("ximagebinsize")->realValue()),
		   yimagebinsize(p->parameter("yimagebinsize")->realValue());
	int	   ximagesize(p->parameter("ximagesize")->intValue()),
		   yimagesize(p->parameter("yimagesize")->intValue());
	bool   withxranges(p->parameter("withxranges")->boolValue()),
		   withyranges(p->parameter("withyranges")->boolValue()),
		   squarepixels(p->parameter("squarepixels")->boolValue());
	double ximagemin(p->parameter("ximagemin")->realValue()),
		   yimagemin(p->parameter("yimagemin")->realValue()),
		   ximagemax(p->parameter("ximagemax")->realValue()),
		   yimagemax(p->parameter("yimagemax")->realValue());

	Info::Task *ev = task("evselect");
	ev->setParameter("keepfilteroutput",false);
	ev->setParameter("withfilteredset",false);
	ev->setParameter("filteredset","");
	ev->setParameter("expression","true");
	ev->setParameter("writedss", false);
	ev->setParameter("updateexposure", false);
	ev->setParameter("withimageset", true);
	ev->setParameter("imagebinning", imagebinning);
	ev->setParameter("squarepixels", squarepixels);
	ev->setParameter("xcolumn", xcolumn);
	ev->setParameter("ximagebinsize", ximagebinsize);
	ev->setParameter("ximagesize", ximagesize);
	ev->setParameter("withxranges", withxranges);
	ev->setParameter("ximagemin", ximagemin);
	ev->setParameter("ximagemax", ximagemax);
	ev->setParameter("ycolumn", ycolumn);
	ev->setParameter("yimagebinsize", yimagebinsize);
	ev->setParameter("yimagesize", yimagesize);
	ev->setParameter("withyranges", withyranges);
	ev->setParameter("yimagemin", yimagemin);
	ev->setParameter("yimagemax", yimagemax);
	ev->setParameter("withimagedatatype", true);
	ev->setParameter("imagedatatype", "Real64");

	for (unsigned n=0; n<colNames.size(); ++n) {
		string imgName(tmpnam(0));
		ev->setParameter("imageset", imgName);
		ev->setParameter("withzcolumn", true);
		ev->setParameter("zcolumn", colNames[n]);
		ev->setParameter("table", _tempTableName);
		errstr<<"Constructing "<<colNames[n]<<" image with eveselect ..."
			  <<Sas::message(Msg::App, Msg::Verbose);
		ev->exec();

		Array *arr = dataSetServer->open(imgName, Dal::Read)->array(0);
		ArrayData *ad = arr->data();
		unsigned Nx = ad->size()[0],
				 Ny = ad->size()[1];
		real64 *adata = ad->real64Data();
		if (!colImage)
			colImage = new ColorImage(Nx, Ny);
		longVec posVec(2);
		for (unsigned nx=0; nx<Nx; ++nx)
			for (unsigned ny=0; ny<Ny; ++ny) {
				posVec[0] = nx; posVec[1] = ny;
				(*colImage)(nx, ny).component(n) = adata[ad->position(posVec)];
			}

		delete ad;

		dataSetServer->close(arr->parent());
		dataSetServer->remove(imgName);
	}

	return colImage;
} // TableBasedPseudoRGBImage::toColorImage()

//
//	convert to a ColorImage using evselect (Pedro Rodriguez procedure)
//
ColorImage *TableBasedPseudoRGBImage::toTrueColorImage() const
{
	ColorImage *colImage = 0;

	vector<string> colNames(3);
	colNames[0] = "RED";
	colNames[1] = "GREEN";
	colNames[2] = "BLUE";

	ParameterInfo *p = Info::Task::current()->parameters();
	string xcolumn("X"),
		   ycolumn("Y"),
		   imagebinning("binSize");
	double ximagebinsize(p->parameter("ximagebinsize")->realValue()),
		   yimagebinsize(p->parameter("yimagebinsize")->realValue());
	bool   withxranges(p->parameter("withxranges")->boolValue()),
		   withyranges(p->parameter("withyranges")->boolValue());
	double ximagemin(p->parameter("ximagemin")->realValue()),
		   yimagemin(p->parameter("yimagemin")->realValue()),
		   ximagemax(p->parameter("ximagemax")->realValue()),
		   yimagemax(p->parameter("yimagemax")->realValue());

	Info::Task *ev = task("evselect");
	ev->setParameter("keepfilteroutput",false);
	ev->setParameter("withfilteredset",false);
	ev->setParameter("filteredset","");
	ev->setParameter("expression","true");
	ev->setParameter("writedss", false);
	ev->setParameter("updateexposure", false);
	ev->setParameter("withimageset", true);
	ev->setParameter("imagebinning", imagebinning);
	ev->setParameter("xcolumn", xcolumn);
	ev->setParameter("ximagebinsize", ximagebinsize);
	ev->setParameter("withxranges", withxranges);
	ev->setParameter("ximagemin", ximagemin);
	ev->setParameter("ximagemax", ximagemax);
	ev->setParameter("ycolumn", ycolumn);
	ev->setParameter("yimagebinsize", yimagebinsize);
	ev->setParameter("withyranges", withyranges);
	ev->setParameter("yimagemin", yimagemin);
	ev->setParameter("yimagemax", yimagemax);
	ev->setParameter("withimagedatatype", true);
	ev->setParameter("imagedatatype", "Real64");

	//etruecolor assume that the input event files have the 
	//same astrometry origin. This fix has been introduced after 
	//runnning etruecolor with mosaicing images.
	//ev->setParameter("raimagecenter",_ra_pnt);
	//ev->setParameter("decimagecenter",_dec_pnt);

	for (unsigned n=0; n<colNames.size(); ++n) {
		string imgName(tmpnam(0));
		ev->setParameter("imageset", imgName);
		ev->setParameter("withzcolumn", true);
		ev->setParameter("zcolumn", colNames[n]);
		ev->setParameter("table", _tempTableName);
		errstr<<"Constructing "<<colNames[n]<<" image with eveselect ..."
			  <<Sas::message(Msg::App, Msg::Verbose);
		ev->exec();

		Array *arr = dataSetServer->open(imgName, Dal::Read)->array(0);
		ArrayData *ad = arr->data();
		unsigned Nx = ad->size()[0],
				 Ny = ad->size()[1];
		//Be aware that the X is the Y and viceversa....
		//With respect to the FITS ARRAY
		real64 *adata = ad->real64Data();
		if (!colImage)
			colImage = new ColorImage(Nx, Ny);
		longVec posVec(2);
		for (unsigned nx=0; nx<Nx; ++nx)
			for (unsigned ny=0; ny<Ny; ++ny) {
				posVec[0] = nx; posVec[1] = ny;
				(*colImage)(nx, ny).component(n) = adata[ad->position(posVec)]/_livetime;
			}

		delete ad;

		dataSetServer->close(arr->parent());
		dataSetServer->remove(imgName);
	}

	return colImage;
} // TableBasedPseudoRGBImage::toColorImage()

TableBasedPseudoRGBImage::~TableBasedPseudoRGBImage()
{
	dataSetServer->remove(_tempTableName);
} // TableBasedPseudoRGBImage::~TableBasedPseudoRGBImage()
