//-----------------------------------------------------------------------------
//
//                      XMM Science Analysis System (SAS)
//                     (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	  etruecolor.cc	--- EPIC true color images
//
//	DESCRIPTION:
//		Construct RGB image data cube.
//
//	CHANGE HISTORY:
//		AI, 2009-04-30: 1.0.0: - first version
//
//	AUTHORS:
//  Aitor Ibarra Ibaibarriaga (AII)
//
//


#include <iostream>
#include <fstream>
#include "Dal.h"
#include "metatask.h"
#include "common.h"
#include "SasFileLocator.h"	//	for locating d/f color tables
#include "stringutil.h"
#include "Color.h"
#include "ColorTable.h"
#include "TableBasedPseudoRGBImage.h"

#include "etruecolor.h"
using namespace std;


void etruecolor()
{
	ParameterInfo *p = Info::Task::current()->parameters();

	//	construct color curve
	string colorTableSetName;

	//	locate color table with file locator

	string ctname = p->parameter("colortable")->stringValue();
	colorTableSetName =
		fileLocator.locate(string("lib/data/colortables/")+ctname + ".ds",
				string("Could not locate color table `") + ctname + "'");

	ColorTable colorTab(colorTableSetName,false);

	//input files
	vector<string> evlistFileName = stringListParameter("tablelist");

	//	construct temporary data set with R/G/B columns
	double minVal = 0.,
		   maxVal = 0.;
	TableBasedPseudoRGBImage::MinMaxDeterminationScheme scheme =
		TableBasedPseudoRGBImage::EXPLICIT;
	minVal = p->parameter("min")->realValue();
	maxVal = p->parameter("max")->realValue();

	//
	Table *table =
					dataSetServer->open(colorTableSetName, Dal::Read)->table("COLORTABLE");
	int N = table->rows();
	ColumnData *pnWeightC = table->column("EPN")->data(),
			*m1WeightC = table->column("EMOS1")->data(),
			*m2WeightC = table->column("EMOS2")->data();

	real32	*pnWeightD = pnWeightC->real32Data(),
			*m1WeightD = m1WeightC->real32Data(),
			*m2WeightD = m2WeightC->real32Data();

	// Add the Color images and weights
	vector<double> pnWeight,m1Weight,m2Weight;
	for (int i=0; i < N ; ++i) pnWeight.push_back(pnWeightD[i]);

	for (int i=0; i < N ; ++i) m1Weight.push_back(m1WeightD[i]);

	for (int i=0; i < N ; ++i) m2Weight.push_back(m2WeightD[i]);



	delete pnWeightC;
	delete m1WeightC;
	delete m2WeightC;

	dataSetServer->close(table->parent());
	//Weights

	//Extract the ObsID information an close file
	Table *inTab = dataSetServer->open(evlistFileName[0], Dal::Read)->table(evlistFileName[0]);
	string obsID = inTab->attribute("OBS_ID")->asString();
	dataSetServer->close(inTab->parent());
	vector<ColorImage*> colImageVec;
	int exposure;

	for (unsigned i = 0 ; i < evlistFileName.size() ; ++i)
	{
		//Filter table for events below 300 eV and above 7000 eV
		Info::Task *evs = task("evselect");

		errstr<<"Filtering dataset " <<Sas::message(Msg::App, Msg::Verbose);

		string oriname = evlistFileName[i];
		oriname.replace(evlistFileName[i].find("."),3,"_evig.ds");
		evs->setParameter("table",evlistFileName[i]);
		evs->setParameter("keepfilteroutput",true);
		evs->setParameter("withfilteredset",true);
		evs->setParameter("filteredset", oriname);
		stringstream expression;
		expression << "(PI in ["<<minVal <<":" << maxVal<<"]) ";
		evs->setParameter("expression",expression.str());
		evs->setParameter("withzcolumn",false);
		evs->setParameter("withimageset",true);
		std::string imgName(tmpnam(0));
		evs->setParameter("imageset",imgName);
		evs->setParameter("xcolumn","X");
		evs->setParameter("ycolumn","Y");
		evs->exec();
		evs->release();

		//Run evigweight to create a WEIGHT column
		//Check if we already have a WEIGHT column created with evigweght
		DataSet *set = dataSetServer->open(evlistFileName[i], Dal::Read);
		if (set->hasAttribute("WITHEFFE"))
			errstr<<"evigweight already executed... " <<Sas::message(Msg::App, Msg::Verbose);
		else
		{
			Info::Task *evi = task("evigweight");
			evi->setParameter("ineventset",oriname);
			evi->setParameter("newoutput",false);
			evi->exec();
		}
		string instr;
		if (set->hasAttribute("INSTRUME"))
			instr = set->attribute("INSTRUME")->asString();
		else
			errstr << "Unknown instrument"<< Sas::error("instrumentnotfound");
		string filter;
		if (set->hasAttribute("FILTER"))
			filter = set->attribute("FILTER")->asString();
		else
			errstr << "Unknown filter"<< Sas::error("filternotfound");


		dataSetServer->close(set);

		TableBasedPseudoRGBImage rgbPseudoImage(
								oriname,
								"X","Y","PI",
								scheme, minVal, maxVal,	colorTab);

		//	construct real color image
		ColorImage *rgbImage = rgbPseudoImage.toTrueColorImage();
		colImageVec.push_back(rgbImage);

		//Set instrument
		colImageVec[i]->setInstrument(instr);
		colImageVec[i]->setFilter(filter);


		Array *imArr = dataSetServer->open(imgName, Dal::Read)->array(0);
		exposure = imArr->attribute("EXPOSURE")->asInt();
		stringstream expo ;
		expo << exposure;
		colImageVec[i]->setExposure(expo.str());
		dataSetServer->remove(imgName);
		dataSetServer->remove(oriname);
	}



	//Add and weight images
	//if there is only 1 input file we dont have to apply any weight
	for(unsigned int i = 0; i < colImageVec.size() ;  ++i)
	{
		// Add the Color images and weights
		vector<double> weight;
		if (trim(colImageVec[i]->getInstrument()) == "EPN")
		{
			weight = pnWeight;
		}
		else if (trim(colImageVec[i]->getInstrument()) == "EMOS1")
		{
			weight = m1Weight;
		}
		else
		{
			weight = m2Weight;
		}

		if (i != 0)
		{
			colImageVec[0]->add(colImageVec[i],weight);
			colImageVec[0]->setInstrument(colImageVec[i]->getInstrument());
			colImageVec[0]->setFilter(colImageVec[i]->getFilter());
			colImageVec[0]->setExposure(colImageVec[i]->getExposure());
		}
		else
			colImageVec[0]->applyWeight(weight);


	}



	//Save the true color images in rates.
	colImageVec[0]->setTrueColor(true);
	//colImageVec[0]->setExposure(expo);

	stringstream rateFileName;
	if (p->parameter("fileset")->stringValue() == "default")
	{
		rateFileName << "P"<<obsID<<"EPX000RGBIMA0000.FIT";
		colImageVec[0]->writeDataSet(rateFileName.str(),
				stringListParameter("tablelist"),true,
			    realParameter("scalmin"),
			    realParameter("scalmax"));
	}
	else
		colImageVec[0]->writeDataSet(p->parameter("fileset")->stringValue(),
				stringListParameter("tablelist"),true,
				realParameter("scalmin"),
				realParameter("scalmax"));



	// Normalize color set
	//	log-scaling/normalizing
	colImageVec[0]->logScale(p->parameter("ximagebinsize")->realValue(),p->parameter("scalmin")->realValue(),
				p->parameter("scalmax")->realValue());

	stringstream colorFileName;
	if (p->parameter("outputchoice")->stringValue() == "dataset")
	{
		if (p->parameter("colorset")->stringValue() == "default")
		{
			colorFileName << "P"<<obsID<<"EPX000RGBCOL0000.FIT";
			colImageVec[0]->writeDataSet(colorFileName.str(),
					stringListParameter("tablelist"),
					p->parameter("ascube")->boolValue(),
					realParameter("scalmin"),
					realParameter("scalmax"));
		}
		else
			colImageVec[0]->writeDataSet(p->parameter("colorset")->stringValue(),
								stringListParameter("tablelist"),
								p->parameter("ascube")->boolValue(),
								realParameter("scalmin"),
								realParameter("scalmax"));
	}
	else {
		string ppmName = p->parameter("ppmfile")->stringValue();
		ostream *os = &cout;
		ofstream ofs;

		if (uppercase(ppmName) != "STDOUT") {
			colorFileName << "P"<<obsID<<"EPX000RGBIMA0000.PPM";

			ofs.open(colorFileName.str().c_str());
			os = &ofs;
		}
		colImageVec[0]->writePPM(*os);
	}

} // xcolorcod()
