//-----------------------------------------------------------------------------
//
//                      XMM Science Analysis System (SAS)
//                     (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	  xcolorcod.cc	--- color coding of spectral images
//
//	DESCRIPTION:
//		Construct RGB image data cube.
//
//	CHANGE HISTORY:
//		MB, 1999-??-??: 1.0.0: - first version
//		UL, 2000-07-17: 2.0.0: - general overhaul in the process of fixing
//								 SPR 855
//							   - is now a metatask that invokes evselect()
//								 to do the image creation
//		UL, 2001-04-04: 2.1.0: - added parameter "ppmfile"
//
//	AUTHORS:
//		Marco Beijersbergen (MB)
//		Uwe Lammers (UL)
//
// $Header: /sasbuild/repositories/sasdev/sas/packages/xcolorcod/src/xcolorcod.cc,v 1.13 2013/08/02 12:09:09 aibarra Exp $
//

#include <iostream>
#include <fstream>
#include "Dal.h"
#include "metatask.h"
#include "common.h"
#include "SasFileLocator.h"	//	for locating d/f color tables
#include "stringutil.h"
#include "Color.h"
#include "ColorTable.h"
#include "TableBasedPseudoRGBImage.h"

#include "xcolorcod.h"

using namespace std;

void xcolorcod()
{
	ParameterInfo *p = Info::Task::current()->parameters();

	//	construct color curve
	string colorTableSetName;
	if (p->parameter("colortablechoice")->stringValue() == "internal") {
		//	locate color table with file locator
		string ctname = p->parameter("colortable")->stringValue();
		colorTableSetName =
			fileLocator.locate(string("lib/data/colortables/")+ctname + ".ds",
					string("Could not locate color table `") + ctname + "'");
	} else
		//	name given explicitely
		colorTableSetName = p->parameter("colortableset")->stringValue();

	ColorTable colorTab(colorTableSetName,
						p->parameter("invertcolortable")->boolValue());

	//	construct temporary data set with R/G/B columns
	string minMaxScheme = p->parameter("ecolumnminmax")->stringValue();
	double minVal = 0.,
		   maxVal = 0.;
	TableBasedPseudoRGBImage::MinMaxDeterminationScheme scheme =
										TableBasedPseudoRGBImage::ACTUAL;
	if (minMaxScheme == "legalminmax")
		scheme = TableBasedPseudoRGBImage::LEGAL;
	else if (minMaxScheme == "explicitminmax") {
		scheme = TableBasedPseudoRGBImage::EXPLICIT;
		minVal = p->parameter("min")->realValue();
		maxVal = p->parameter("max")->realValue();
	}



	TableBasedPseudoRGBImage rgbPseudoImage(
								p->parameter("table")->stringValue(),
								p->parameter("xcolumn")->stringValue(),
								p->parameter("ycolumn")->stringValue(),
								p->parameter("ecolumn")->stringValue(),
								scheme, minVal, maxVal,	colorTab);

	//	construct real color image
	ColorImage *rgbImage = rgbPseudoImage.toColorImage();

	//Input file name. Needed to write WCS info in the output file
	std::vector<std::string> ev;
	ev.push_back(p->parameter("table")->stringValue());

	//	log-scaling/normalizing
	if (uppercase(p->parameter("scale")->stringValue()) == "LOG")
		rgbImage->log(p->parameter("decades")->realValue());
	rgbImage->normalize(255);
	if (p->parameter("outputchoice")->stringValue() == "dataset")
		rgbImage->writeDataSet(p->parameter("colorset")->stringValue(), ev,
							   p->parameter("ascube")->boolValue());
	else {
		string ppmName = p->parameter("ppmfile")->stringValue();
		ostream *os = &cout;
		ofstream ofs;
		if (uppercase(ppmName) != "STDOUT") {
			ofs.open(ppmName.c_str());
			os = &ofs;
		}
		rgbImage->writePPM(*os);
	}
} // xcolorcod()
